"""
(ngs)xfem
=========

A module for unfitted discretizations in NGSolve

Modules:
xfem.lsetcuving ... isoparametric unfitted FEM
"""

from ngsolve.comp import *
from ngsolve.fem import *
from ngsolve import BitArray
from ngsolve.utils import L2
from xfem.ngsxfem_py import *
# from xfem.ngsxfem_utils_py import *
# from xfem.ngsxfem_lsetcurving_py import *
# from xfem.ngsxfem_xfem_py import *
# from xfem.ngsxfem_cutint_py import *
# from xfem.ngsxfem_spacetime_py import *

def extend(func):
    """
Evaluates the XFiniteElement-function independent of the level set domains.

Note:
This will lead to the same behavior as the function that the XFiniteElement-function is based
on.
    """
    if func.derivname == "extend":
        return func.Deriv()
    add = func.Operator("extend")
    if add:
        return add
    raise Exception("cannot form extend")

def pos(func):
    """
Evaluates an XFiniteElement-function assuming a positive level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-positive.
    """
    if func.derivname == "pos":
        return func.Deriv()
    add = func.Operator("pos")
    if add:
        return add
    raise Exception("cannot form pos")

def neg(func):
    """
Evaluates an XFiniteElement-function assuming a negative level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-negative.
    """
    if func.derivname == "neg":
        return func.Deriv()
    add = func.Operator("neg")
    if add:
        return add
    raise Exception("cannot form neg")

def extend_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function independent of the level set domains.

Note:
This will lead to the same behavior as the function that the XFiniteElement-function is based on.
    """
    if func.derivname == "extendgrad":
        return func.Deriv()
    add = func.Operator("extendgrad")
    if add:
        return add
    raise Exception("cannot form extend_grad")

def pos_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function assuming a positive level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-positive.
    """
    if func.derivname == "posgrad":
        return func.Deriv()
    add = func.Operator("posgrad")
    if add:
        return add
    raise Exception("cannot form pos_grad")

def neg_grad(func):
    """
Evaluates the gradient of an XFiniteElement-function assuming a negative level set domain.

Note:
This can lead to non-zero values also in domains where the level set function is non-negative.
    """
    if func.derivname == "neggrad":
        return func.Deriv()
    add = func.Operator("neggrad")
    if add:
        return add
    raise Exception("cannot form neg_grad")

SymbolicBFI_old = SymbolicBFI
def SymbolicBFI(levelset_domain=None, *args, **kwargs):
    """
Wrapper around SymbolicBFI to allow for integrators on level set domains (see also
SymbolicCutBFI). The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
SymbolicBFI function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": ngsolve.CoefficientFunction
    CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
    FESpace with scalar continuous piecewise (multi-) linear basis functions.
  * "domain_type" : {NEG,POS,IF} (ENUM)
    Integration on the domain where either:
    * the level set function is negative (NEG)
    * the level set function is positive (POS)
    * the level set function is zero     (IF )
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "force_intorder" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

Other Parameters :

  form : ngsolve.CoefficientFunction
    form to integrate

  VOL_or_BND : {VOL,BND}
    integrator is defined in the volume or boundary

  element_boundary : boolean
    Integration of the boundary of an element
    (not active for level set domains)

  skeleton : boolean
    Integration over element-interface

  definedon : Region
    Domain description on where the integrator is defined

  definedonelements: BitArray
    BitArray that allows integration only on elements or facets (if skeleton=True) that are marked
    True.

  deformation : GridFunction
    Specify a specific mesh deformation for a bilinear form

  time_order : int
    order in time that is used in the space-time integration. time_order=-1 means that no space-time
    rule will be applied. This is only relevant for space-time discretizations.
"""
    if levelset_domain != None and type(levelset_domain)==dict:
        if not "force_intorder" in levelset_domain:
            levelset_domain["force_intorder"] = -1
        if not "subdivlvl" in levelset_domain:
            levelset_domain["subdivlvl"] = 0
        if not "levelset" in levelset_domain:
            print("Please provide a level set function")
        if not "domain_type" in levelset_domain:
            print("Please provide a domain type (NEG,POS or IF)")
        if not "quad_dir_policy" in levelset_domain:
            levelset_domain["quad_dir_policy"] = OPTIMAL
        # print("SymbolicBFI-Wrapper: SymbolicCutBFI called")
        return SymbolicCutBFI(lset=levelset_domain["levelset"],
                              domain_type=levelset_domain["domain_type"],
                              force_intorder=levelset_domain["force_intorder"],
                              subdivlvl=levelset_domain["subdivlvl"],
                              quad_dir_policy=levelset_domain["quad_dir_policy"],
                              *args, **kwargs)
    else:
        # print("SymbolicBFI-Wrapper: original SymbolicBFI called")
        if (levelset_domain == None):
            return SymbolicBFI_old(*args,**kwargs)
        else:
            return SymbolicBFI_old(levelset_domain,*args,**kwargs)

SymbolicLFI_old = SymbolicLFI
def SymbolicLFI(levelset_domain=None, *args, **kwargs):
    """
Wrapper around SymbolicLFI to allow for integrators on level set domains (see also
SymbolicCutLFI). The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
SymbolicLFI function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": ngsolve.CoefficientFunction
    CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
    FESpace with scalar continuous piecewise (multi-) linear basis functions.
  * "domain_type" : {NEG,POS,IF} (ENUM)
    Integration on the domain where either:
    * the level set function is negative (NEG)
    * the level set function is positive (POS)
    * the level set function is zero     (IF )
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "force_intorder" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

Other Parameters :

  form : ngsolve.CoefficientFunction
    form to integrate

  VOL_or_BND : {VOL,BND}
    integrator is defined in the volume or boundary

  element_boundary : boolean
    Integration of the boundary of an element
    (not active for level set domains)

  skeleton : boolean
    Integration over element-interface

  definedon : Region
    Domain description on where the integrator is defined

  definedonelements: BitArray
    BitArray that allows integration only on elements or facets (if skeleton=True) that are marked
    True.

  deformation : GridFunction
      Specify a specific mesh deformation for a linear form

  time_order : int
    order in time that is used in the space-time integration. time_order=-1 means that no space-time
    rule will be applied. This is only relevant for space-time discretizations.
"""
    if levelset_domain != None and type(levelset_domain)==dict:
        if not "force_intorder" in levelset_domain:
            levelset_domain["force_intorder"] = -1
        if not "subdivlvl" in levelset_domain:
            levelset_domain["subdivlvl"] = 0
        if not "levelset" in levelset_domain:
            print("Please provide a level set function")
        if not "domain_type" in levelset_domain:
            print("Please provide a domain type (NEG,POS or IF)")
        if not "quad_dir_policy" in levelset_domain:
            levelset_domain["quad_dir_policy"] = OPTIMAL
        # print("SymbolicLFI-Wrapper: SymbolicCutLFI called")
        return SymbolicCutLFI(lset=levelset_domain["levelset"],
                              domain_type=levelset_domain["domain_type"],
                              force_intorder=levelset_domain["force_intorder"],
                              subdivlvl=levelset_domain["subdivlvl"],
                              quad_dir_policy=levelset_domain["quad_dir_policy"],
                              *args, **kwargs)
    else:
        # print("SymbolicLFI-Wrapper: original SymbolicLFI called")
        if (levelset_domain == None):
            return SymbolicLFI_old(*args,**kwargs)
        else:
            return SymbolicLFI_old(levelset_domain,*args,**kwargs)

def Integrate_X_special_args(levelset_domain={}, cf=None, mesh=None, VOL_or_BND=VOL, order=5, time_order=-1, region_wise=False, element_wise = False, heapsize=1000000):
    """
Integrate_X_special_args should not be called directly.
See documentation of Integrate.
    """
    if not "force_intorder" in levelset_domain or levelset_domain["force_intorder"] == -1:
        levelset_domain["force_intorder"] = -1
    else:
        order = levelset_domain["force_intorder"]

    if not "subdivlvl" in levelset_domain:
        levelset_domain["subdivlvl"] = 0
    if not "levelset" in levelset_domain:
        print("Please provide a level set function")
    if not "domain_type" in levelset_domain:
        print("Please provide a domain type (NEG,POS or IF)")
    if not "quad_dir_policy" in levelset_domain:
        levelset_domain["quad_dir_policy"] = OPTIMAL

    return IntegrateX(lset=levelset_domain["levelset"],
                      mesh=mesh, cf=cf,
                      order=order,
                      domain_type=levelset_domain["domain_type"],
                      subdivlvl=levelset_domain["subdivlvl"],
                      time_order=time_order,
                      quad_dir_policy=levelset_domain["quad_dir_policy"],
                      heapsize=heapsize)


##### THIS IS ANOTHER WRAPPER (original IntegrateX-interface is pretty ugly...) TODO
Integrate_old = Integrate
def Integrate(levelset_domain=None, *args, **kwargs):
    """
Integrate-wrapper. If a dictionary 'levelset_domain' is provided integration will be done on the
level set part of the mesh. The dictionary contains the level set function (CoefficientFunciton or
GridFunction) and the domain-type (NEG/POS/IF). If the dictionary is not provided, the standard
Integrate function from NGSolve will be called.

Parameters

levelset_domain : dictionary
  entries:
  * "levelset": ngsolve.CoefficientFunction
    CoefficientFunction that describes the geometry. In the best case lset is a GridFunction of an
    FESpace with scalar continuous piecewise (multi-) linear basis functions.
  * "domain_type" : {NEG,POS,IF} (ENUM)
    Integration on the domain where either:
    * the level set function is negative (NEG)
    * the level set function is positive (POS)
    * the level set function is zero     (IF )
  * "subdivlvl" : int
    On simplex meshes a subtriangulation is created on which the level set function lset is
    interpolated piecewise linearly. Based on this approximation, the integration rule is
    constructed. Note: this argument only works on simplices.
  * "force_intorder" : int
    (default: entry does not exist or value -1)
    overwrites "order"-arguments in the integration (affects only spatial integration)
  * "time_order" : int 
    defines integration order in time (for space-time integrals only)
  * "quad_dir_policy" : {FIRST, OPTIMAL, FALLBACK} (ENUM)
    Integration direction policy for iterated integrals approach
    * first direction is used unless not applicable (FIRST)
    * best direction (in terms of transformation constant) is used (OPTIMAL)
    * subdivision into simplices is always used (FALLBACK)

mesh :
  Mesh to integrate on (on some part)

cf : ngsolve.CoefficientFunction
  the integrand

order : int (default = 5)
  integration order. Can be overruled by "force_intorder"-entry of the levelset_domain dictionary.

time_order : int (default = -1)
  integration order in time (for space-time integration), default: -1 (no space-time integrals)

region_wise : bool
  (only active for non-levelset version)

element_wise : bool
  (only active for non-levelset version)

heapsize : int
  heapsize for local computations.
    """
    if levelset_domain != None and type(levelset_domain)==dict:
        # print("Integrate-Wrapper: IntegrateX called")
        return Integrate_X_special_args(levelset_domain, *args, **kwargs)
    else:
        # print("Integrate-Wrapper: original Integrate called")
        if (levelset_domain == None):
            return Integrate_old(*args,**kwargs)
        else:
            newargs = [levelset_domain]
            for q in args:
                newargs.append(q)
            return Integrate_old(*newargs,**kwargs)

def IndicatorCF(mesh, ba, facets = False):
    """
Returns a CoefficientFunction that evaluates a BitArray. On elements/facets with an index i where
the BitArray evaluates to true the CoefficientFunction will evaluate as 1, otherwise as 0. Similar
functionality (only on elements) can be obtained with BitArrayCF.
    """
    if facets:
        ret = GridFunction(FESpace("facet",mesh,order=0))
        for i in range(len(ba)):
            if ba[i]:
                ret.vec[i] = 1.0
            else:
                ret.vec[i] = 0.0
        return ret
    else:
        return BitArrayCF(BitArray(ba))

def CutRatioGF(cutinfo):
    """
Ratio between negative and full part of an element. Vector taken from CutInfo and put into a
piecewise constant GridFunction.
    """
    ret = GridFunction(L2(cutinfo.Mesh(),order=0))
    ret.vec.data = cutinfo.GetCutRatios(VOL)
    return ret

def kappa(mesh,lset_approx, subdivlvl=0):
    """
Tuple of ratios between negative/positive and full
part of an element (deprecated).
    """
    print("kappa-function is deprecated - use CutRatioGF instead")
    kappa1 = GridFunction(L2(mesh,order=0))
    lset_neg = { "levelset" : lset_approx, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    kappa_f = LinearForm(kappa1.space)
    kappa_f += SymbolicLFI(levelset_domain = lset_neg, form = kappa1.space.TestFunction() )
    kappa_f.Assemble();
    kappa1.space.SolveM(CoefficientFunction(1.0),kappa_f.vec)
    kappa1.vec.data = kappa_f.vec
    kappa2 = 1.0 - kappa1
    return (kappa1,kappa2)

def IsCut(mesh,lset_approx, subdivlvl=0):
    """
GridFunction that is 1 on cut elements, 0 otherwise (deprecated). Use CutInfo-functionality (perhaps
combined with BitArrayCF).
    """
    print("IsCut-function is deprecated - use CutInfo-functionality instead")
    def cut(kappa):
        if (kappa > 1e-16 and kappa < 1.0-1e-16):
            return 1.0
        else:
            return 0.0
    kappa1, dummycf = kappa(mesh,lset_approx,subdivlvl)
    for i in range(len(kappa1.vec)):
        kappa1.vec[i] = cut(kappa1.vec[i])
    # from numpy import vectorize
    # cut = vectorize(cut)
    # kappa1.vec.FV().NumPy()[:] = cut(kappa1.vec.FV().NumPy())
    return kappa1

def SpaceTimeWeakSet(gfu_e, cf, space_fes):
    gfu_e_repl = GridFunction(space_fes)
    gfu_e_repl.Set( cf )
    gfu_e.vec[:].data = gfu_e_repl.vec



from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from xfem import *
from numpy import pi


class LevelSetMeshAdaptation_Spacetime:
    """
    Deformation from level set in space-time based on the 
    LevelSetMeshAdaptation class.
    """

    order_deform = 2
    order_qn = 2
    order_lset = 2

    def __init__(self, mesh, order_space = 2, order_time = 1, lset_lower_bound = 0,
                 lset_upper_bound = 0, threshold = -1, discontinuous_qn = False, heapsize=1000000,periodic=False):
        """
        Deformation
        """
        self.order_deform = order_space
        self.order_qn = order_space
        self.order_lset = order_space
        self.order_time = order_time

        self.lset_lower_bound = lset_lower_bound
        self.lset_upper_bound = lset_upper_bound
        self.threshold = threshold
        self.periodic = periodic
        
        self.v_ho = H1(mesh, order=self.order_lset)
        self.lset_ho_node = GridFunction (self.v_ho, "lset_ho_node")
        self.ndof_node = len(self.lset_ho_node.vec)

        if (discontinuous_qn):
            self.v_qn = L2(mesh, order=self.order_qn, dim=mesh.dim)
        else:
            self.v_qn = H1(mesh, order=self.order_qn, dim=mesh.dim)
        self.qn = GridFunction(self.v_qn, "qn")
    
        self.v_p1 = H1(mesh, order=1)
        self.lset_p1_node = GridFunction (self.v_p1, "lset_p1_node")
        self.ndof_node_p1 = len(self.lset_p1_node.vec)

        if self.periodic:
            self.v_def = Periodic(H1(mesh, order=self.order_deform, dim=mesh.dim))
        else:
            self.v_def = H1(mesh, order=self.order_deform, dim=mesh.dim)
        
        self.deform_node = GridFunction(self.v_def, "deform_node")
        self.heapsize = heapsize
        
        # Spacetime
        self.tfe = ScalarTimeFE(self.order_time) 
        
        self.v_ho_st = SpaceTimeFESpace(self.v_ho,self.tfe)
        self.lset_ho = GridFunction(self.v_ho_st)
        
        self.v_p1_st = SpaceTimeFESpace(self.v_p1,self.tfe)
        self.lset_p1 = GridFunction(self.v_p1_st)
        
        self.v_def_st = SpaceTimeFESpace(self.v_def,self.tfe)
        self.deform = GridFunction(self.v_def_st)
        
        self.ci = CutInfo(mesh)
        
        self.hasneg_spacetime = BitArray(self.ci.GetElementsOfType(NEG))
        self.hasneg_spacetime[:] = False
        self.haspos_spacetime = BitArray(self.ci.GetElementsOfType(POS))
        self.haspos_spacetime[:] = False
        self.hasif_spacetime = BitArray(self.ci.GetElementsOfType(IF))
        self.hasif_spacetime[:] = False
        
        self.v_kappa_node = L2(mesh,order=0)
        self.v_kappa = SpaceTimeFESpace(self.v_kappa_node,self.tfe)
        self.kappa = GridFunction(self.v_kappa, "kappa")        
        
    def interpol_ho(self,levelset,t): #,tstart,delta_t):
        #times = [tstart + delta_t * xi for xi in self.v_ho_st.TimeFE_nodes()]
        times = [xi for xi in self.v_ho_st.TimeFE_nodes()]
        for i,ti in enumerate(times):
            #t.Set(ti)
            t.FixTime(ti)
            #print("i, ti: ", i, ti)
            self.lset_ho_node.Set(levelset)
            self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node].data = self.lset_ho_node.vec[:]
        #t.Set(tstart)
        t.UnfixTime()

    def interpol_p1(self):
        for i in range(self.order_time + 1):
            self.lset_ho_node.vec[:].data = self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node]
            InterpolateToP1(self.lset_ho_node,self.lset_p1_node)
            self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1].data = self.lset_p1_node.vec[:]
            
    def CalcDeformation(self, levelset,t,calc_kappa = False):
        """
        Compute the deformation
        """
        self.v_ho.Update()
        self.lset_ho_node.Update()
        self.v_p1.Update()
        self.lset_p1_node.Update()
        self.v_qn.Update()
        self.qn.Update()
        self.v_def.Update()
        self.deform_node.Update()
        self.v_kappa_node.Update()
        self.v_kappa.Update()
        self.kappa.Update()
        
        self.interpol_ho(levelset,t) #,tstart,delta_t)
        self.interpol_p1()
                
        for i in  range(len(self.v_ho_st.TimeFE_nodes())):
            self.lset_p1_node.vec[:].data = self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1]
            self.ci.Update(self.lset_p1_node)
            if calc_kappa:
                self.kappa.vec[i*self.v_kappa_node.ndof : (i+1)*self.v_kappa_node.ndof].data = self.ci.GetCutRatios(VOL)
        
        
        self.ci.Update(self.lset_p1,self.order_time)
        self.hasneg_spacetime[:] = False
        self.hasneg_spacetime |= self.ci.GetElementsOfType(NEG)
        self.haspos_spacetime[:] = False
        self.haspos_spacetime |= self.ci.GetElementsOfType(POS)
        self.hasif_spacetime[:] = False
        self.hasif_spacetime |= self.ci.GetElementsOfType(IF)
        
        
        for i in range(self.order_time + 1):
            self.lset_ho_node.vec[:].data = self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node]
            self.qn.Set(self.lset_ho_node.Deriv())
            self.lset_p1_node.vec[:].data = self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1]
            ProjectShift(self.lset_ho_node, self.lset_p1_node, self.deform_node, self.qn, self.hasif_spacetime, None, self.lset_lower_bound, 
                         self.lset_upper_bound, self.threshold, heapsize=self.heapsize)
            self.deform.vec[i*self.ndof_node : (i+1)*self.ndof_node].data = self.deform_node.vec[:]
        return self.deform
            
    def CalcMaxDistance(self, levelset,t, given_pts = []):
        """
        Compute largest distance
        """
        if given_pts:
            time_quad = given_pts
        else:
            time_quad = self.v_ho_st.TimeFE_nodes() 
        # times = [tstart + delta_t * xi for xi in time_quad]
        max_dists = []
        # for ti,xi in zip(times,time_quad):
        for xi in time_quad:
        
            RestrictGFInTime(self.lset_p1,xi,self.lset_p1_node)
            RestrictGFInTime(self.deform,xi,self.deform_node)
            t.FixTime(xi)
            # t.Set(ti) 
            self.v_def_st.SetTime(xi)
            self.v_ho_st.SetTime(xi)
            max_dists.append(CalcMaxDistance(levelset,self.lset_p1_node,self.deform_node,heapsize=self.heapsize))
            #max_dists.append(CalcMaxDistance(self.lset_ho,self.lset_p1,self.deform,heapsize=self.heapsize))
        t.UnfixTime()
        self.v_def_st.SetOverrideTime(False)
        self.v_ho_st.SetOverrideTime(False)
        # t.Set(tstart)
        return max(max_dists)
      
       
## geometry        
#square = SplineGeometry()
#square.AddRectangle([0,0],[2,2],bc=1)
#ngmesh = square.GenerateMesh(maxh=0.03, quad_dominated=False)
#mesh = Mesh (ngmesh)
#
## data
#t = Parameter(0)
#lset = CoefficientFunction( sqrt( (x-1-0.25*sin(pi*t))*(x-1-0.25*sin(pi*t))+(y-1)*(y-1)) - 0.5  )
#
#
#
#def SolveProblem(mesh,delta_t,k_s=2,k_t=1):
#    max_nodes = []
#    max_interm = []
#    tstart = 0
#    tnew = 0
#    tend = 1
#    t.Set(tnew)
#    lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = k_t,
#                                             threshold=0.5, discontinuous_qn=True)
#    
#    while tend - tnew > delta_t/2:
#        t.Set(tnew)
#        dfm = lset_adap_st.CalcDeformation(lset,t,tnew,delta_t,calc_kappa=True) 
#        Draw(lset_adap_st.kappa,mesh,"kappa")
#        max_nodes.append(lset_adap_st.CalcMaxDistance(lset,t,tnew,delta_t))
#        max_interm.append(lset_adap_st.CalcMaxDistance(
#                              lset,t,tnew,delta_t,[i*0.1 for i in range(11)]))
#        tnew += delta_t 
#    return max(max_nodes),max(max_interm)
#
#
#def StudyConvergence(delta_t=0.5,max_rfs=2,where = "space"):   
#    max_dist_nodes = []
#    max_dist_interm = []
#    ref_lvl = 0
#    while ref_lvl <= max_rfs:  
#        
#        if where == "space" and ref_lvl > 0:
#            mesh.Refine()
#        elif where == "time" and ref_lvl > 0:
#            delta_t = delta_t / 2  
#        e1,e2 = SolveProblem(mesh,delta_t,k_s=3,k_t=2)
#        max_dist_nodes.append(e1)
#        max_dist_interm.append(e2)
#        ref_lvl = ref_lvl + 1
#    print("Studying convergence in: " + where)
#    print("Max-Dist nodes = {0}".format(max_dist_nodes))
#    if min(max_dist_nodes):
#        eoc_nodes = [ log(max_dist_nodes[i-1]/max_dist_nodes[i])/log(2) for i in range(1,len(max_dist_nodes))]
#        print("Eoc nodes = {0}".format(eoc_nodes))
#    print("Max-Dist intermediate = {0}".format(max_dist_interm))
#    if min(max_dist_interm):
#        eoc_interm = [ log(max_dist_interm[i-1]/max_dist_interm[i])/log(2) for i in range(1,len(max_dist_interm))]
#        print("Eoc intermediate = {0}".format(eoc_interm))
#            
#            
#StudyConvergence(delta_t=0.01,max_rfs=3,where = "time")
#        
#                                            
#lset_adap_st.interpol_ho(lset,t,tstart,delta_t)
#lset_ho = lset_adap_st.lset_ho
#lset_adap_st.interpol_p1()
#lset_p1 = lset_adap_st.lset_p1
#dfm = lset_adap_st.CalcDeformation(lset,t,tstart,delta_t)
#print("Max-Dist nodes = {0}".format(lset_adap_st.CalcMaxDistance(lset,t,tstart,delta_t)))
#print("Max-Dist intermediate points = {0}".format(lset_adap_st.CalcMaxDistance(
#                              lset,t,tstart,delta_t,[i*0.1 for i in range(11)])))
#
## Plotting
#visoptions.deformation = 1
#lset_adap_st.v_ho_st.SetTime(0.0)    
#lset_adap_st.v_p1_st.SetTime(0.0) 
#lset_adap_st.v_def_st.SetTime(0.0) 
#Draw(lset_ho,mesh,"lsetHO")
#Draw(lset_p1,mesh,"lsetP1")
#Draw(dfm,mesh,"deformation")
#input("")
#lset_adap_st.v_ho_st.SetTime(1.0) 
#lset_adap_st.v_p1_st.SetTime(1.0)  
#lset_adap_st.v_def_st.SetTime(1.0)  
#Redraw()   




try:
    import skbuild
except ImportError:
    print("")
    print("***********************************************************************************")
    print("")
    print("skbuild not found, please execute:")
    print("pip3 install scikit-build")
    print("")
    print("***********************************************************************************")
    print("")

from skbuild import setup

CMAKE_ARGS = [ "-DBUILD_NGSOLVE=OFF" ]

# CMAKE_ARGS.append( "-DCMAKE_OSX_DEPLOYMENT_TARGET=10.9" )
# CMAKE_ARGS.append( "-DCMAKE_OSX_SYSROOT=/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk" )

setup(name="xfem",
      version="0.9.3",
      description="",
      packages=["xfem", "xfem.lsetcurv", "xfem.lset_spacetime", "xfem.utils"],
      package_dir={"xfem" : "python",
                   "xfem.lsetcurv" : "lsetcurving",
                   "xfem.lset_spacetime" : "spacetime",
                   "xfem.utils" : "utils"},
      cmake_args=CMAKE_ARGS)

